/*
 * Implementation of the abstract base class.
 *
 * We export the TclObj_proc procedure, to which all objects point.
 * It is called whenever an object is ``invoked''. Its task is to
 * look up a matching signature and then invoke the correct handler
 * for that member.
 *
 * The Tcl_Enable member renders an object visible from the Tcl
 * level (by assigning a name to the object and entering it into
 * the command table). It is usually invoked silently whenever an
 * object is used as return value from the new operator or from
 * TCL_OBJECT_RETURN_OBJECT.
 *
 * The base class provides the "info" member on the Tcl level via
 * the Tcl_Proc member function. The function is inherited by all
 * derived classes.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <ctype.h>
#include "tclobj.h"
#include "compat.h"
#include "signature.h"
#include "register.h"
#include "fptools.h"


/*
 * cvs version
 */
char * tclobj_tclobj_id = "$Id: tclobj.cc,v 1.3 1996/11/28 15:12:13 fp Exp $";


/*
 * ----------------------------------------------------------------------
 * Local Initializer, Tcl_Proc function and object info
 * ----------------------------------------------------------------------
 */

extern "C" {
int
Tcl_Object_Tcl_Proc (Tcl_Object *myself, Tcl_Interp *interp,
		     int argc, char *argv[], int command)
{
  return myself->Tcl_Proc (interp, argc, argv, command);
}
/*
 * This function is called from initializers of derived classes.
 */
int
C_Tcl_Object_Init (Tcl_Interp *interp)
{
  extern int Tclobj_Init (Tcl_Interp *);
  return Tclobj_Init (interp);
}
}

static const TclObj_Comtab Tcl_Object_mytable[] = {
  { 1, "info", "...", "get information about object and its operators" },
  { 0, NULL, NULL, NULL }
};

TclObj_Info C_Tcl_Object_Info = {
  0, "Tcl_Object", NULL, Tcl_Object_mytable, NULL, NULL,
  0, NULL, NULL, Tcl_Object_Tcl_Proc
};

TclObj_Info * Tcl_Object::Tcl_Object_myinfo = &C_Tcl_Object_Info;

/*
 * ======================================================================
 * Main Entry Point: This function is called whenever one of the objects
 *                   is "executed". After identifying the object, check
 *                   its command table and see if the command is valid
 *                   and if the parameters are correct.
 * ======================================================================
 */

extern "C" {
static int
TclObj_proc (ClientData clientData, Tcl_Interp *interp, 
	     int argc, char *argv[])
{
  Tcl_Object *myself = (Tcl_Object *) clientData;
  const TclObj_Argtree *thetree;
  char tmpstr[256], *token;
  int index, cmdfound=0;
  Tcl_CmdInfo cmdinfo;

  if (myself == NULL) {
    sprintf (tmpstr, "%s: oops, what am i?", argv[0]);
    Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
    return TCL_ERROR;
  }

  if (argc < 2) {
    sprintf (tmpstr, "%s: object of type %s at address %lu",
	     argv[0], myself->classinfo()->name, (unsigned long) myself);
    Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
    return TCL_OK;
  }

  cmdfound = TCL_OK;
  thetree  = TclObj_SearchSignature (interp, myself->classinfo(),
				     argv[1], argc-2, argv+2,
				     &cmdfound);

  if (cmdfound == TCL_ERROR) {
    /*
     * for example: ilegal scope resolution operator
     */
    return TCL_ERROR;
  }

  if (thetree == NULL) {
    if (TclObj_DumpMemberNames (interp, myself->classinfo(),
				argv[1], &cmdfound) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if (!cmdfound) {
      /*
       * list all available operators (for this scope)
       */
      Tcl_AppendResult (interp,
			"bad action \"", argv[1], "\": should be one of ",
			NULL);

      if (strchr (argv[1], ':') == NULL) {
	strcpy (token=tmpstr, argv[1]);
	while (isalnum(*token)) token++;
	while (*token==':')     token++;
	*token   = '\0';
	cmdfound = 0;
	if (TclObj_DumpMemberNames (interp, myself->classinfo(),
				    tmpstr, &cmdfound) == TCL_ERROR)
	  return TCL_ERROR;

	if (cmdfound) {
	  TclObj_DumpMemberNames (interp, myself->classinfo(),
				  tmpstr, NULL);
	}
	else {
	  TclObj_DumpMemberNames (interp, myself->classinfo(),
				  NULL, NULL);
	}
      }
      else {
	TclObj_DumpMemberNames (interp, myself->classinfo(),
				NULL, NULL);
      }
    }
    else {
      /*
       * we had a command of that name, but the signature didn't match.
       * make a list of all available signature and explain why it didn't
       * match.
       */
      Tcl_AppendResult (interp, "bad signature: \"", argv[1], NULL);
      for (index=2; index<argc; index++) {
	if (Tcl_GetCommandInfo (interp, argv[index], &cmdinfo)) {
	  if (cmdinfo.clientData) {
	    Tcl_AppendResult (interp, " ",
			      ((Tcl_Object*)cmdinfo.clientData)->classinfo()->name,
			      NULL);
	  }
	  else {
	    /* huh - does not appear to be an object */
	    Tcl_AppendResult (interp, " *", NULL);
	  }
	}
	else {
	  /* just a plain string */
	  Tcl_AppendResult (interp, " *", NULL);
	}
      }

      Tcl_AppendResult      (interp, "\" does not match any of\n", NULL);
      TclObj_DumpSignatures (interp, myself->classinfo(), argv[1], NULL);
    }

    return TCL_ERROR;
  }

  /*
   * okay, back to the real world again. the arguments have
   * checked out okay, so fire up the procedure!
   */

  return thetree->provider->c_tcl_proc (myself, interp, argc, argv,
					thetree->entry->number);
}
} /* extern "C" */

/* 
 * ======================================================================
 * Base Class Implementation
 * ======================================================================
 */

/*
 * Make us accessible from the Tcl level
 */

const char *
Tcl_Object::Tcl_Enable (Tcl_Interp * interp,
			const char * thename)
{
  char newname[256], tmpstr[256];
  const char *theclass;
  unsigned long counter;

  theclass = classinfo()->name;

  if (theclass==ourclasstype && thename==NULL) {
    return Tcl_Name (interp);
  }

  /*
   * If we already exist, disable us first. Maybe we are being
   * redefined from a derived class
   */

  Tcl_Disable ();

  /*
   * do not generate us if the derived object doesn't want us to
   */

  if (theclass == NULL)
    return NULL;

  /*
   * if thename==NULL, generate name for new object
   */

  if (thename == NULL) {
    thename = newname;
    counter = classinfo()->counter += 1;
    sprintf (newname, "%s%lu", theclass, counter);
  }

  /*
   * enter myself into reverse mapping table
   */

  sprintf     (tmpstr, "%lu", (unsigned long) this);
  Tcl_SetVar2 (interp, "TclObj_rev", tmpstr, (char *) thename,
	       TCL_GLOBAL_ONLY);

  /*
   * Create myself as command
   */

  Tcl_CreateCommand (interp, (char *) thename, TclObj_proc,
		     (ClientData) this, NULL);

  ourclasstype = theclass;

  return thename;
}

/*
 * Common Destructor. Remove myself as command and all the table
 * references
 */

void
Tcl_Object::Tcl_Disable (void)
{
  const char * ptrname;
  char tmpstr[32];

  /*
   * if we've been just a phantom
   */
  if (ourclasstype == NULL || interpreter == NULL) {
    return;
  }

  /*
   * who am i? do not use Tcl_Name, as that would cause a
   * disabled object to be enabled first.
   */

  sprintf (tmpstr, "%lu", (unsigned long) this);
  if ((ptrname = Tcl_GetVar2 (interpreter, "TclObj_rev", tmpstr,
			      TCL_GLOBAL_ONLY)) == NULL) {
    /* oops */
    Tcl_AppendResult (interpreter, "Tcl_Object::Tcl_Disable(",
		      ourclasstype, "): who am I?", NULL);
    ourclasstype = NULL;
    return;
  }

  /*
   * Delete myself as command
   */

  Tcl_DeleteCommand (interpreter, (char *) ptrname);

  /*
   * Remove myself from reverse mapping table
   */

  sprintf       (tmpstr, "%lu", (unsigned long) this);
  Tcl_UnsetVar2 (interpreter, "TclObj_rev", (char *) tmpstr, TCL_GLOBAL_ONLY);

  /*
   * Suicide complete
   */
  ourclasstype = NULL;
}

/*
 * Retrieve our command name
 */

const char *
Tcl_Object::Tcl_Name (Tcl_Interp *interp)
{
  char tmpstr[32], *thename;

  sprintf (tmpstr, "%lu", (unsigned long) this);
  if ((thename = Tcl_GetVar2 (interp, "TclObj_rev", tmpstr,
			      TCL_GLOBAL_ONLY)) == NULL) {
    return Tcl_Enable (interp);
  }
  return thename;
}

/*
 * Tcl_Proc function. Handle some common commands, otherwise
 * produce an error message.
 */

int
Tcl_Object::Tcl_Proc (Tcl_Interp *interp,
		      int argc, char *argv[], int command)
{
  char tmpstr[256];
  int index;

  /*
   * provide the "info" command
   */

  if (command != 1) {
    Tcl_AppendResult (interp, argv[0],
		    "(", classinfo()->name, ")",
		    ": lost in Tcl_Object::Tcl_Proc",
		    NULL);
    return TCL_ERROR;
  }

  if (argc == 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"",
		      argv[0], " info {id|dep|op|sig|<func>}\"",
		      NULL);
    return TCL_OK;
  }

  if ((TclObj_DumpMemberNames (interp,classinfo(),argv[2],&index))==TCL_ERROR)
    return TCL_ERROR;

  if (index) {
    return TclObj_DumpSignatures (interp, classinfo(), argv[2], NULL);
  }
  else if (_FP_strnicmp (argv[2], "id", 2) == 0) {
    sprintf (tmpstr, "\"%s\" is an object of type \"%s\" at address %lu",
	     argv[0], classinfo()->name, (unsigned long) this);
    Tcl_SetResult (interp, tmpstr, TCL_VOLATILE);
    return TCL_OK;
  }
  else if (_FP_strnicmp (argv[2], "dep", 3) == 0) {
    return TclObj_DumpComptree (interp, classinfo());
  }
  else if (_FP_strnicmp (argv[2], "op", 2) == 0) {
    return TclObj_DumpMemberNames (interp, classinfo(), NULL, NULL);
  }
  else if (_FP_strnicmp (argv[2], "sig", 2) == 0) {
    return TclObj_DumpSignatures (interp, classinfo(), NULL, NULL);
  }
  else {
    Tcl_AppendResult (interp, argv[0],
		      " does not support ",
		      argv[2], NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}
